///////////////////////////////////////////////////////////////////////////////
// big_if_statement.hpp
//
// Concepts:
//  IfStatement is...
//    a type which has a static member function, 
//    bool compile(Params);
//
// Note:
//  big_if_statement can't touch Params type,
//  so you should choose whether referenced or not.
//

#ifndef KETCHUP_BIG_IF_STATEMENT_H
#define KETCHUP_BIG_IF_STATEMENT_H

#ifdef _MSC_VER
#pragma once
#endif

#include <boost/concept_check.hpp>
#include <boost/mpl/reverse_fold.hpp>

namespace ketchup {

	namespace detail {

	///////////////////////////////////////////////////////////////////////////////
	// Note:
	//  If IfStatementSeq has the same IfStatement, VC7.1 duplicates
	//  the static member function, IfStatement::compile (LNK1179 error).
	//  I can't understand the problem,
	//  but VC7.1 seems to hate static member functions.
	//  So I define fold_op_workaround and...
	//
	// compile_if_statement
	//
	template< class IfStatement, class Params >
	bool compile_if_statement(Params params)
	{
		return IfStatement::compile(params);
	}

	} // namespace detail
	
///////////////////////////////////////////////////////////////////////////////
// big_if_statement
//
template< class IfStatementSeq, class Params >
struct big_if_statement
{
	struct fold_init_state
	{
		static bool compile(Params params)
		{
			boost::ignore_unused_variable_warning(params);
			return false;
		}
	};
	
	template< class State, class IfStatement >
	struct fold_op
	{
		static bool compile(Params params)
		{
			if (detail::compile_if_statement<IfStatement, Params>(params))
				return true;
			else
				return State::compile(params);
		}
	};

	// The placeholders say mpl_assert_not_arg, so I define
	struct fold_op_workaround
	{
		template< class State, class IfStatement >
		struct apply
		{
			static bool compile(Params params)
			{
				if (detail::compile_if_statement<IfStatement, Params>(params))
					return true;
				else
					return State::compile(params);
			}

			typedef apply type;
		};
	};

	typedef typename boost::mpl::reverse_fold<IfStatementSeq,
		fold_init_state,
		// fold_op<boost::mpl::_1, boost::mpl::_2> // makes mpl_assert_not_arg
		fold_op_workaround
	>::type type;
};

} // namespace ketchup

#endif
